import abc


class ZmqClientError(Exception):
    pass


class ZmqClient(metaclass=abc.ABCMeta):

    def __init__(self, context, server_addr):
        self._server_addr = server_addr
        self._context = context

    @abc.abstractmethod
    def _connect(self):
        pass

    @abc.abstractmethod
    def _disconnect(self):
        pass



import gpiozero
from .exceptions import PinError


class GpioZeroManager:

    def __init__(self, pub_socket):
        self.pub_socket = pub_socket
        self.pins = {}

    def _lookup_pin(self, pin):
        try:
            return self.pins[pin]
        except KeyError:
            raise PinError('Pin not active')

    def _pin_pub_callback(self, event):
        def callback(device):
            message = {
                'pin': device.pin.number,
                'event': event
            }
            self.pub_socket.send_json(obj=message)
        return callback

    def add_input(self, pin, pull_up=False, bounce_time=None):
        self.pins[pin] = gpiozero.DigitalInputDevice(pin, pull_up=pull_up, bounce_time=bounce_time)

    def add_output(self, pin, active_high=True, initial_value=False):
        self.pins[pin] = gpiozero.DigitalOutputDevice(pin, active_high=active_high, initial_value=initial_value)

    def pin_read(self, pin):
        pin = self._lookup_pin(pin=pin)
        return pin.value

    def pin_on(self, pin):
        pin = self._lookup_pin(pin=pin)
        pin.on()
        return pin.value

    def pin_off(self, pin):
        pin = self._lookup_pin(pin=pin)
        pin.off()
        return pin.value

    def enable_pub_when_activated(self, pin):
        pin = self._lookup_pin(pin=pin)
        pin.when_activated = self._pin_pub_callback(event='activated')

    def enable_pub_when_deactivated(self, pin):
        pin = self._lookup_pin(pin=pin)
        pin.when_deactivated = self._pin_pub_callback(event='deactivated')

    def disable_pub_when_activated(self, pin):
        pin = self._lookup_pin(pin=pin)
        pin.when_activated = None

    def disable_pub_when_deactivated(self, pin):
        pin = self._lookup_pin(pin=pin)
        pin.when_deactivated = None

    def clean_up(self):
        for _, pin in self.pins.items():
            pin.close()



from jsonrpc import JSONRPCResponseManager
from jsonrpc.dispatcher import Dispatcher
import zmq
from pi_gpio_core.gpio import GpioZeroManager


class GpioCoreServer:

    def __init__(self, rpc_socket, gpio_manager):
        self.rpc_socket = rpc_socket
        self.gpio_manager = gpio_manager
        self.dispatcher = Dispatcher()
        self.dispatcher.add_method(self.gpio_manager.add_input)
        self.dispatcher.add_method(self.gpio_manager.add_output)
        self.dispatcher.add_method(self.gpio_manager.pin_read)
        self.dispatcher.add_method(self.gpio_manager.pin_on)
        self.dispatcher.add_method(self.gpio_manager.pin_off)
        self.dispatcher.add_method(self.gpio_manager.enable_pub_when_activated)
        self.dispatcher.add_method(self.gpio_manager.disable_pub_when_activated)
        self.dispatcher.add_method(self.gpio_manager.enable_pub_when_deactivated)
        self.dispatcher.add_method(self.gpio_manager.disable_pub_when_deactivated)

    def run(self):
        try:
            while True:
                message = self.rpc_socket.recv_string()
                response = JSONRPCResponseManager.handle(message, self.dispatcher)
                self.rpc_socket.send_string(response.json)
        finally:
            self.gpio_manager.clean_up()


def gpio_core_server(rpc_port=5555, pub_port=5556):
    context = zmq.Context()
    rpc_socket = context.socket(zmq.REP)
    rpc_socket.bind('tcp://127.0.0.1:{0}'.format(rpc_port))
    pub_socket = context.socket(zmq.PUB)
    pub_socket.bind('tcp://127.0.0.1:{0}'.format(pub_port))
    gpio_manager = GpioZeroManager(pub_socket=pub_socket)
    return GpioCoreServer(port=rpc_socket, gpio_manager=gpio_manager)



from .reqrep_client import ReqRepClient
from .sub_client import SubClient


class GpioCoreClientError(Exception):

    def __init__(self, message, **kwargs):
        self.__dict__.update(kwargs)
        super().__init__(message)


class GpioCoreRpcClient(ReqRepClient):

    def __init__(self, port=5555):
        super().__init__(server_addr='tcp://127.0.0.1:{0}'.format(port))
        self._last_id = None

    def _next_payload_id(self):
        if self._last_id is None:
            self._last_id = 0
        else:
            self._last_id += 1
        return self._last_id

    def request(self, method, params):
        payload = {
            'id': self._next_payload_id(),
            'method': method,
            'params': params,
            'jsonrpc': '2.0'
        }
        resp = super().request(obj=payload)
        try:
            return resp['result']
        except KeyError:
            resp['error'].pop('message')
            raise GpioCoreClientError(resp['error']['data']['message'], **resp['error'])

    def add_input(self, pin, pull_up=False, bounce_time=None):
        return self.request(method='add_input', params=[pin, pull_up, bounce_time])

    def add_output(self, pin):
        return self.request(method='add_output', params=[pin])

    def pin_read(self, pin):
        return self.request(method='pin_read', params=[pin])

    def pin_on(self, pin):
        return self.request(method='pin_on', params=[pin])

    def pin_off(self, pin):
        return self.request(method='pin_off', params=[pin])

    def enable_pub_when_activated(self, pin):
        return self.request(method='enable_pub_when_activated', params=[pin])

    def disable_pub_when_activated(self, pin):
        return self.request(method='disable_pub_when_activated', params=[pin])

    def enable_pub_when_deactivated(self, pin):
        return self.request(method='enable_pub_when_deactivated', params=[pin])

    def disable_pub_when_deactivated(self, pin):
        return self.request(method='disable_pub_when_deactivated', params=[pin])


class GpioCoreSubClient(SubClient):

    def __init__(self, port=5556):
        super().__init__(server_addr='tcp://127.0.0.1:{0}'.format(port))



import zmq
from .zmq_client import ZmqClient


class SubClient(ZmqClient):

    def __init__(self, server_addr):
        context = zmq.Context()
        super().__init__(context=context, server_addr=server_addr)

    def _connect(self):
        self._socket = self._context.socket(zmq.SUB)
        self._socket.connect(self._server_addr)

    def _disconnect(self):
        self._socket.setsockopt(zmq.LINGER, 0)
        self._socket.close()

    def messages(self):
        yield self._socket.recv_json()



from distutils.core import setup

setup(
    name='pi-gpio-core',
    version='0.0.10',
    author='Brian Hines',
    author_email='brian@projectweekend.net',
    packages=['pi_gpio_core'],
    package_data={
        'pi_gpio_core': [
            'client/*',
            'server/*'
        ]
    },
    url='https://github.com/exitcodezero/pi-gpio-core',
    license='LICENSE',
    description='A ZeroMQ-based service for interacting with GPIO pins',
    long_description='A ZeroMQ-based service for interacting with GPIO pins',
    install_requires=[
        "coverage==4.0.3",
        "gpiozero==1.1.0",
        "json-rpc==1.10.3",
        "nose==1.3.7",
        "pyzmq==15.2.0",
        "RPi.GPIO==0.6.2"
    ],
)



import json
import zmq
from .zmq_client import ZmqClient, ZmqClientError


class ReqRepClient(ZmqClient):

    def __init__(self, server_addr, max_retries=3, timeout=3000):
        context = zmq.Context()
        super().__init__(context=context, server_addr=server_addr)
        self.max_retries = max_retries
        self.remaining_retries = self.max_retries
        self.timeout = timeout
        self._connect()

    def _connect(self):
        self._socket = self._context.socket(zmq.REQ)
        self._socket.connect(self._server_addr)
        self._poll = zmq.Poller()
        self._poll.register(self._socket, zmq.POLLIN)

    def _disconnect(self):
        self._socket.setsockopt(zmq.LINGER, 0)
        self._socket.close()
        self._poll.unregister(self._socket)

    def request(self, obj):
        while self.remaining_retries:
            json_data = json.dumps(obj=obj)
            self._socket.send_string(json_data)
            expect_reply = True
            while expect_reply:
                socks = dict(self._poll.poll(self.timeout))
                if socks.get(self._socket) == zmq.POLLIN:
                    response = self._socket.recv_string()
                    if not response:
                        break
                    else:
                        self.remaining_retries = self.max_retries
                        expect_reply = False
                        return json.loads(response)
                else:
                    self._disconnect()
                    self.remaining_retries -= 1
                    if self.remaining_retries == 0:
                        # server is not online so stop trying
                        break
                    self._connect()
                    self._socket.send_string(json_data)
        else:
            self._context.term()
            raise ZmqClientError('Max reconnect attempts exceeded')



import unittest
import gpiozero
from .gpio import GpioZeroManager


class GpioTestCase(unittest.TestCase):

    def setUp(self):
        super(GpioTestCase, self).setUp()
        self.gpio_manager = GpioZeroManager(gpio_lib=gpiozero)

    def test_add_input(self):
        self.gpio_manager.add_input(pin=10)
        self.assertIsInstance(self.gpio_manager.pins[10], gpiozero.DigitalInputDevice)
